"""
Some fixtures are mocked to avoid having to connect to a real database.
A mysql database is available while running tests locally using docker-compose.
The database is not available when running tests on github actions.
The envionment variable GITHUB_ACTIONS is set to true when running on github actions.
https://docs.github.com/en/actions/learn-github-actions/variables#default-environment-variables
"""


import os
from typing import Callable, Generator

import pytest
import yaml
from blackline.adapters.mysql.mysql import MySQLAdapter
from blackline.factories.query import QueryFactory
from blackline.models.mysql.mysql import MySQLConfig
from blackline.utils.testing.conftest_shared import *  # noqa: F403, F401
from mysql.connector import MySQLConnection
from mysql.connector.abstracts import MySQLConnectionAbstract
from mysql.connector.cursor import MySQLCursor
from pytest import MonkeyPatch


@pytest.fixture
def github_actions_mysql() -> bool:
    if os.getenv("GITHUB_ACTIONS") == "true":
        """This will help testing the monkeypatches locally.
        If set to true, the connections to mysql will fail if not correctly mocked.
        """
        return True
    return False


@pytest.fixture
def mysql_store_name() -> str:
    return "test_mysql"


@pytest.fixture
def mysql_user() -> str:
    return os.environ.get("MYSQL_USER", "no_user_set")


@pytest.fixture
def mysql_password() -> str:
    return os.environ.get("MYSQL_PASSWORD", "no_password_set")


@pytest.fixture
def mysql_host() -> str:
    return os.environ.get("MYSQL_HOST", "127.0.0.1")


@pytest.fixture
def mysql_port() -> int:
    if os.getenv("GITHUB_ACTIONS") == "true":
        """This will help testing the monkeypatches locally.
        If set to true, the connections to mysql will fail if not correctly mocked.
        """
        return 5555
    return int(os.environ.get("MYSQL_PORT", 3306))


@pytest.fixture
def mysql_db() -> str:
    return os.environ.get("MYSQL_DB", "no_db_set")


@pytest.fixture
def stores_yaml(
    mysql_store_name: str,
    mysql_user: str,
    mysql_password: str,
    mysql_host: str,
    mysql_port: int,
    mysql_db: str,
) -> str:
    """https://www.mysqlql.org/docs/current/libpq-connect.html#LIBPQ-PARAMKEYWORDS"""
    return f"""
    name: {mysql_store_name}
    profiles:
      dev:
        type: mysql
        config:
          connection:
            user: {mysql_user}
            password: {mysql_password}
            database: {mysql_db}
            host: {mysql_host}
            port: {mysql_port}
        """


@pytest.fixture
def mysql_connection(
    mysql_user: str,
    mysql_password: str,
    mysql_host: str,
    mysql_port: str,
    mysql_db: str,
    github_actions_mysql: bool,
    monkeypatch: MonkeyPatch,
) -> MySQLConnection:
    if github_actions_mysql:

        def __init__(self, **kwargs):
            self._consume_results = False
            self._unread_result = False
            if kwargs:
                self.connect(**kwargs)

        def cursor(self, *args, **kwargs) -> MySQLCursor:
            return MySQLCursor(self)

        monkeypatch.setattr(MySQLConnection, "__init__", __init__)
        monkeypatch.setattr(MySQLConnection, "close", lambda self: None)
        monkeypatch.setattr(MySQLConnection, "cursor", cursor)
        monkeypatch.setattr(MySQLConnection, "commit", lambda self: None)
        monkeypatch.setattr(MySQLConnection, "is_connected", lambda self: True)
        monkeypatch.setattr(
            MySQLConnectionAbstract, "connect", lambda self, **kwargs: None
        )
        monkeypatch.setattr(MySQLCursor, "execute", lambda self, *args, **kwargs: None)
        monkeypatch.setattr(
            MySQLCursor, "executemany", lambda self, *args, **kwargs: None
        )

    return MySQLConnection(
        user=mysql_user,
        password=mysql_password,
        host=mysql_host,
        port=mysql_port,
        database=mysql_db,
    )


@pytest.fixture
def load_database(
    mysql_connection: MySQLConnection, mock_data: list, test_table: str
) -> Generator[MySQLConnectionAbstract, None, None]:
    with mysql_connection as conn:
        with conn.cursor() as cursor:
            cursor.execute(f"DROP TABLE IF EXISTS {test_table}")
            cursor.execute(
                f"""
                CREATE TABLE {test_table} (
                    id CHAR(36) DEFAULT (UUID()) UNIQUE,
                    created_at TIMESTAMP,
                    name VARCHAR(255),
                    email VARCHAR(255) NOT NULL,
                    postal_code VARCHAR(15) UNIQUE,
                    active BOOLEAN,
                    ip VARCHAR(15),
                    CHECK (created_at > '2020-01-01 00:00:00'),
                    PRIMARY KEY (`id`)
                    )"""
            )
            cursor.executemany(
                f"INSERT INTO {test_table} (created_at, name, email, postal_code, active, ip) VALUES (%s, %s, %s, %s, %s, %s)",  # noqa: E501
                mock_data,
            )
            conn.commit()

            yield conn

            cursor.execute(f"DROP TABLE IF EXISTS {test_table}")


@pytest.fixture
def mysql_adapter(
    stores_yaml: str,
    mysql_connection: MySQLConnection,
) -> MySQLAdapter:
    store_obj = yaml.safe_load(stores_yaml)
    mysql_obj = store_obj["profiles"]["dev"]
    pg_config = MySQLConfig.parse_obj(mysql_obj)
    return MySQLAdapter(config=pg_config.config)


@pytest.fixture
def mysql_query_factory(
    query_factory_factory: Callable,
    mysql_adapter: MySQLAdapter,
) -> QueryFactory:
    return query_factory_factory(adapter=mysql_adapter)



import yaml
from blackline.models.mysql.mysql import MySQLConfig
from pydantic.types import SecretStr


def test_MySQLConfig(
    stores_yaml: str,
    mysql_user: str,
    mysql_password: SecretStr,
    mysql_db: str,
    mysql_host: str,
    mysql_port: int,
) -> None:
    # Setup
    store_obj = yaml.safe_load(stores_yaml)
    mysql_obj = store_obj["profiles"]["dev"]

    # Run
    mysql_config = MySQLConfig.parse_obj(mysql_obj)

    # Assert
    assert mysql_config.type == "mysql"
    assert mysql_config.config.connection.host == mysql_host
    assert mysql_config.config.connection.port == mysql_port
    assert mysql_config.config.connection.database == mysql_db
    assert mysql_config.config.connection.user == mysql_user
    assert mysql_config.config.connection.password.get_secret_value() == mysql_password



"""
MySQL adapter configuration. This shouldn't be here but will move is
later and have it dynamically loaded from blackline-postgres.
"""

from typing import Literal, Optional

from blackline.models.adapter import AdapterConfig, ConnectionConfig
from pydantic import BaseModel, Field, SecretStr


class MySQLConnectionConfig(ConnectionConfig):
    user: str = Field(
        ..., description="The user name used to authenticate with the MySQL server."
    )
    password: SecretStr = Field(
        ..., description="The password to authenticate the user with the MySQL server."
    )
    password1: Optional[SecretStr] = Field(
        None,
        description="For Multi-Factor Authentication (MFA); password1 is an alias for password.",  # noqa: E501
    )
    password2: Optional[SecretStr] = Field(
        None, description="For Multi-Factor Authentication (MFA)"
    )
    password3: Optional[SecretStr] = Field(
        None, description="For Multi-Factor Authentication (MFA)"
    )
    database: str = Field(
        ...,
        description="The database name to use when connecting with the MySQL server.",
    )
    host: str = Field(
        "127.0.0.1", description="The host name or IP address of the MySQL server."
    )
    port: int = Field(3306, description="The port number of the MySQL server.")
    unix_socket: Optional[str] = Field(
        None, description="The path to a Unix domain socket."
    )

    class Config:
        extra = "allow"


class MySQLConfig(AdapterConfig):
    class Config(BaseModel):
        connection: MySQLConnectionConfig

    type: Literal["mysql"]
    config: Config



import os
from typing import List

import pytest
import yaml
from blackline.factories.query import QueryFactory
from blackline.models.datastores import DataStore
from blackline.models.mysql.mysql import MySQLConfig
from mysql.connector import MySQLConnection


def test_mysql_store_config(stores_yaml: str):
    # Setup
    pg_store_info = yaml.safe_load(stores_yaml)

    # Run
    store_config = DataStore.parse_obj(pg_store_info)

    # Assert
    isinstance(store_config.profiles["dev"], MySQLConfig)


def test_query_factory_mysql_queries(
    mysql_query_factory: QueryFactory,
):
    # Run
    queries = mysql_query_factory.queries()

    # Assert
    assert (
        queries[0].sql
        == "UPDATE test_table\nSET\n  email = %(email_value)s,\n  name = null\nWHERE created_at < %(cutoff)s"  # noqa: E501
    )
    assert (
        queries[1].sql
        == "UPDATE test_table\nSET\n  ip = REGEXP_REPLACE(ip,'[:alnum:]',%(ip_value)s)\nWHERE created_at < %(cutoff)s"  # noqa: E501
    )


@pytest.mark.skipif(
    os.getenv("GITHUB_ACTIONS") == "true",
    reason="Github Actions does not have a local mysql",
)
def test_query_factory_mysql_execution(
    load_database: MySQLConnection,
    mysql_query_factory: QueryFactory,
    test_table: str,
    deidentified_mock_data: List,
):
    # Run
    queries = mysql_query_factory.queries()
    queries[0].execute()
    queries[1].execute()

    # Assert
    with queries[0].adapter.connection() as conn:
        with conn.cursor() as cursor:
            cursor.execute(f"SELECT * FROM {test_table}")
            rows = cursor.fetchall()

    # Remove the UUIS from the rows
    rows = [row[1:] for row in rows]

    # breakpoint()
    assert set(deidentified_mock_data) == set(rows)



from blackline.adapters.mysql.mysql import MySQLAdapter
from mysql.connector import MySQLConnection
from mysql.connector.cursor import MySQLCursor
from pytest import MonkeyPatch


def test_connection(
    stores_yaml: str,
    monkeypatch: MonkeyPatch,
    github_actions_mysql: bool,
    mysql_adapter: MySQLAdapter,
) -> None:
    # Setup

    if github_actions_mysql:
        monkeypatch.setattr(MySQLCursor, "fetchone", lambda self: (1,))

    # Run
    with mysql_adapter.connection() as conn:
        with conn.cursor() as cur:
            cur.execute("SELECT 1")
            result = cur.fetchone()

    # Assert
    assert result is not None
    assert result[0] == 1


def test_is_connected(
    mysql_adapter: MySQLAdapter, monkeypatch: MonkeyPatch, github_actions_mysql: bool
) -> None:
    # Run & Assert
    assert mysql_adapter.is_connected()


def test_columns(
    monkeypatch: MonkeyPatch,
    github_actions_mysql: bool,
    load_database: MySQLConnection,
    test_table: str,
    mysql_adapter: MySQLAdapter,
):
    # Setup
    def _columns_from_table(self, table: str):
        return [
            ("id", "char(36)", "NO", "PRI", "uuid()", "DEFAULT_GENERATED"),
            ("created_at", "timestamp", "YES", "", None, ""),
            ("name", "varchar(255)", "YES", "", None, ""),
            ("email", "varchar(255)", "NO", "", None, ""),
            ("postal_code", "varchar(15)", "YES", "UNI", None, ""),
            ("active", "tinyint(1)", "YES", "", None, ""),
            ("ip", "varchar(15)", "YES", "", None, ""),
        ]

    if github_actions_mysql:
        monkeypatch.setattr(MySQLAdapter, "_columns_from_table", _columns_from_table)
        monkeypatch.setattr(
            MySQLCursor,
            "fetchall",
            lambda self: _columns_from_table(self, test_table),
        )

    # Run
    columns = mysql_adapter.columns(test_table)

    with mysql_adapter.connection() as con:
        with con.cursor() as cur:
            cur.execute(f"SHOW COLUMNS FROM {test_table}")
            result = cur.fetchall()

    # Assert
    for i, column in enumerate(columns):
        assert column.name == result[i][0]
        assert column.data_type == result[i][1]
        assert column.default == result[i][4]

        if result[i][2] == "YES":
            assert column.nullable is True

        if result[i][3] == "PRI":
            assert column.primary_key is True
            assert column.unique is True

        if result[i][3] == "UNI":
            assert column.unique is True


def test_table_exist(
    monkeypatch: MonkeyPatch,
    github_actions_mysql: bool,
    load_database: MySQLConnection,
    test_table: str,
    mysql_adapter: MySQLAdapter,
):
    # Setup
    if github_actions_mysql:
        monkeypatch.setattr(MySQLCursor, "fetchall", lambda self: [(f"{test_table}",)])

    # Run
    exists = mysql_adapter.table_exists(table=test_table)

    assert exists is True


def test_table_exist_does_not_exist(
    monkeypatch: MonkeyPatch,
    github_actions_mysql: bool,
    mysql_adapter: MySQLAdapter,
):
    # Setup
    nonexistent_table = "nonexistent_table"
    if github_actions_mysql:
        monkeypatch.setattr(MySQLCursor, "fetchall", lambda self: [])

    # Run
    exists = mysql_adapter.table_exists(table=nonexistent_table)

    assert exists is False



from contextlib import contextmanager
from typing import Generator, Union

from blackline.adapters.sql.sql import SQLAdapter
from blackline.models.collection import Column
from blackline.models.mysql.mysql import MySQLConfig
from mysql.connector import MySQLConnection


class MySQLAdapter(SQLAdapter):
    def __init__(self, config: MySQLConfig.Config, *args, **kwargs):
        super().__init__(config=config, *args, **kwargs)
        self.database = self.config.connection.database

    @contextmanager
    def connection(self) -> Generator[MySQLConnection, None, None]:
        conn_args = self.config.connection.dict()
        conn_args["password"] = conn_args["password"].get_secret_value()
        conn = MySQLConnection(**conn_args)
        try:
            yield conn
        except Exception as e:
            conn.rollback()
            raise e
        else:
            conn.commit()
        finally:
            conn.close()

    def test_connection(self) -> bool:
        return self.is_connected()

    def is_connected(self) -> bool:
        with self.connection() as conn:
            return conn.is_connected()

    def mask_template(self) -> str:
        return "{{ name }} = REGEXP_REPLACE(ip,'[:alnum:]',%({{ value }})s)"

    def table_exists(self, table: str) -> bool:
        """Check if a table exists.

        Args:
            table (str): Table name.

        Returns:
            bool: True if the table exists.
        """
        with self.connection() as con:
            with con.cursor() as cur:
                cur.execute(
                    f"SHOW TABLES LIKE '{table}'",
                )
                res = cur.fetchall()
        return len(res) > 0

    def columns(self, table: str) -> list[Column]:
        """
        Return a list of columns for a given table.

        Args:
            table: Table name.

        Returns:
            A list of Column.
        """
        columns_from_table = self._columns_from_table(table=table)
        return [
            Column(
                name=column[0],
                data_type=column[1],
                nullable=True if column[2] == "YES" else False,
                unique=True if column[3] == "UNI" or column[3] == "PRI" else False,
                primary_key=True if column[3] == "PRI" else False,
                foreign_key=False,
                check=None,
                default=column[4],
            )
            for column in columns_from_table
        ]

    def _columns_from_table(
        self, table: str
    ) -> list[tuple[str, str, str, str, Union[str, None], str]]:
        with self.connection() as con:
            with con.cursor() as cur:
                cur.execute(f"SHOW COLUMNS FROM {table}")
                res = cur.fetchall()
        return res
